@@ -30,6 +30,8 @@ OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 """
 
+from typing import Union, List, AnyStr, Optional, Dict, Tuple
+
 import autowrap.PXDParser as PXDParser
 import autowrap.Types as Types
 import autowrap.Utils as Utils
@@ -38,7 +40,6 @@ from collections import defaultdict
 from autowrap.tools import OrderKeepingDictionary
 
 
-
 __doc__ = """
 
     The methods in this module take the class declarations created by
@@ -79,7 +80,7 @@ __doc__ = """
         - wrap-manual-memory: will allow the user to provide manual memory
                               management of self.inst, therefore the class will
                               not provide the automated __dealloc__ and inst
-                              attribute (but their presence is still expected). 
+                              attribute (but their presence is still expected).
                               This is useful if you cannot use the shared-ptr
                               approach to store a reference to the C++ class
                               (as with singletons for example).
@@ -110,83 +111,106 @@ __doc__ = """
 
 """
 
-import logging as L
+from autowrap import logger
+
+DEFAULT_NAMESPACE = ""
+CONCURRENT_FILES_PER_CORE = 10
 
 
 class ResolvedTypeDef(object):
-
-    def __init__(self, decl):
-        self.cpp_decl = decl
-        self.name = decl.name
+    def __init__(self, decl: PXDParser.CTypeDefDecl):
+        self.cpp_decl: PXDParser.CTypeDefDecl = decl
+        self.name: str = decl.name
         self.type_ = decl.type_
         self.wrap_ignore = decl.annotations.get("wrap-ignore", False)
+        self.pxd_import_path = None
 
 
 class ResolvedEnum(object):
-
     def __init__(self, decl):
         self.name = decl.name
+        self.scoped = decl.scoped
         self.wrap_ignore = decl.annotations.get("wrap-ignore", False)
         self.cpp_decl = decl
         self.items = decl.items
         self.type_ = Types.CppType(self.name, enum_items=self.items)
-        L.info("created resolved enum: %s" % (decl.name, ))
-        L.info("           with items: %s" % (decl.items))
-        L.info("")
+        self.pxd_import_path = None
+        logger.info("created resolved enum: %s" % (decl.name,))
+        logger.info("           with items: %s" % decl.items)
+        logger.info("")
 
 
 class ResolvedAttribute(object):
-
     def __init__(self, name, type_, decl):
         self.name = name
         self.type_ = type_
         self.cpp_decl = decl
         self.wrap_ignore = decl.annotations.get("wrap-ignore", False)
+        self.pxd_import_path = None
 
 
-default_namespace = ""
-
-def get_namespace(pxd, default_namespace):
+def get_namespace(pxd, default_namespace: str) -> str:
     filehandle = open(pxd)
     fulltext = filehandle.read()
     filehandle.close()
     import re
+
     match = re.search(r"cdef extern.*?namespace\s*\"([^\"]*)\"", fulltext)
     if not match:
         return default_namespace
     else:
         return match.group(1)
 
+
 class ResolvedClass(object):
 
-    """ contains all info for generating wrapping code of
-        resolved class.
-        "Resolved" means that template parameters and typedefs are resolved.
+    """contains all info for generating wrapping code of
+    resolved class.
+    "Resolved" means that template parameters and typedefs are resolved.
     """
 
+    name: str
+    methods: OrderKeepingDictionary
+    attributes: List[ResolvedAttribute]
+    cpp_decl: PXDParser.CppClassDecl
+    ns: AnyStr
+    wrap_ignore: bool
+    no_pxd_import: bool
+    wrap_manual_memory: Union[bool, List[AnyStr]]
+    wrap_hash: List[AnyStr]
+    local_map: Dict
+    instance_map: Dict
+    pxd_import_path: Optional[AnyStr]
+
     def __init__(self, name, methods, attributes, decl, instance_map, local_map):
-        self.name = name
-        # resolve overloadings
-        self.methods = OrderKeepingDictionary()
+        self.name: str = name
+        # resolve overloads
+        self.methods: OrderKeepingDictionary = OrderKeepingDictionary()
         for m in methods:
             self.methods.setdefault(m.name, []).append(m)
         self.attributes = attributes
 
         self.cpp_decl = decl
-        self.ns = get_namespace(decl.pxd_path, default_namespace)
+        self.ns = get_namespace(decl.pxd_path, DEFAULT_NAMESPACE)
 
         self.wrap_ignore = decl.annotations.get("wrap-ignore", False)
         self.no_pxd_import = decl.annotations.get("no-pxd-import", False)
         self.wrap_manual_memory = decl.annotations.get("wrap-manual-memory", [])
-        # fix previous code where we had a bool ... 
+        # fix previous code where we had a bool ...
+        # Note that we need the explicit True and False here since we do not want to overwrite
+        #  non-empty lists that we might get from the line above! And non-empty lists implicitly
+        #  evaluate to True.
+        # TODO test this in test suite or drop backwards compatibility. It has been years.
         if self.wrap_manual_memory == True:
             self.wrap_manual_memory = ["__old-model"]
-        else:
-            self.wrap_manual_memory == []
-        assert( isinstance(self.wrap_manual_memory, list) )
+        elif self.wrap_manual_memory == False:
+            self.wrap_manual_memory = []
+        # elif empty list or list with actual content: pass
+        assert isinstance(self.wrap_manual_memory, list)
         self.wrap_hash = decl.annotations.get("wrap-hash", [])
         self.local_map = local_map
         self.instance_map = instance_map
+        self.pxd_import_path = None
 
     def get_flattened_methods(self):
         return [m for methods in self.methods.values() for m in methods]
@@ -197,21 +221,24 @@ class ResolvedClass(object):
 
 class ResolvedMethod(object):
 
-    """ contains all info for generating wrapping code of
-        resolved class.
-        "resolved" means that template parameters are resolved.
+    """contains all info for generating wrapping code of
+    resolved class.
+    "resolved" means that template parameters are resolved.
     """
 
-    def __init__(self, name, result_type, arguments, decl, instance_map,
-                 local_map):
-        self.name = name
+    def __init__(
+        self, name, is_static, result_type, arguments, decl, instance_map, local_map
+    ):
+        self.name: str = name
+        self.is_static: bool = is_static
         self.result_type = result_type
         self.arguments = arguments
         self.cpp_decl = decl
-        self.wrap_ignore = decl.annotations.get("wrap-ignore", False)
-        self.with_nogil = decl.annotations.get("wrap-with-no-gil", False)
+        self.wrap_ignore: bool = decl.annotations.get("wrap-ignore", False)
+        self.with_nogil: bool = decl.annotations.get("wrap-with-no-gil", False)
         self.local_map = local_map
-        self.instance_amp = instance_map
+        self.instance_map = instance_map
+        self.pxd_import_path = None
 
     def __str__(self):
         args = [("%s %s" % (t, n)).strip() for (n, t) in self.arguments]
@@ -223,22 +250,28 @@ class ResolvedFunction(ResolvedMethod):
     pass
 
 
-def resolve_decls_from_files_single_thread(pathes, root):
+def resolve_decls_from_files(paths, root, num_processes=1, cython_warn_level=1):
+    if num_processes > 1:
+        return resolve_decls_from_files_multi_thread(
+            paths, root, num_processes, cython_warn_level
+        )
+    else:
+        return resolve_decls_from_files_single_thread(paths, root, cython_warn_level)
+
+
+def resolve_decls_from_files_single_thread(paths, root, cython_warn_level=1):
     decls = []
-    for k, path in enumerate(pathes):
+    for k, path in enumerate(paths):
         full_path = os.path.join(root, path)
-        if k % 50 == 0: 
-            L.log(25, "parsing progress %s out of %s" % (k, len(pathes)))
-        decls.extend(PXDParser.parse_pxd_file(full_path))
+        if k % 50 == 0:
+            logger.log(25, "parsing progress %s out of %s" % (k, len(paths)))
+        decls.extend(PXDParser.parse_pxd_file(full_path, cython_warn_level))
     return _resolve_decls(decls)
 
-def resolve_decls_from_files(pathes, root, num_processes = 1):
-    if num_processes > 1:
-        return resolve_decls_from_files_multi_thread(pathes, root, num_processes)
-    else:
-        return resolve_decls_from_files_single_thread(pathes, root)
 
-def resolve_decls_from_files_multi_thread(pathes, root, num_processes):
+def resolve_decls_from_files_multi_thread(
+    paths, root, num_processes, cython_warn_level=1
+):
     """Perform parsing with multiple threads
 
     This function distributes the work on `num_processes` processes and each
@@ -246,23 +279,29 @@ def resolve_decls_from_files_multi_thread(pathes, root, num_processes):
     work on.
     """
     import multiprocessing as mp
+    from functools import partial
 
-    CONCURRENT_FILES_PER_CORE = 10
     pool = mp.Pool(processes=num_processes)
-    full_pathes = [os.path.join(root, path) for path in pathes]
+    full_paths = [os.path.join(root, path) for path in paths]
 
     decls = []
-    while len(full_pathes) > 0:
-        n_work = len(full_pathes)
+    while len(full_paths) > 0:
+        n_work = len(full_paths)
         remaining = max(0, n_work - num_processes * CONCURRENT_FILES_PER_CORE)
-        args = [full_pathes.pop() for k in range(n_work - remaining)]
-        L.log(25, "parsing progress %s out of %s with %s processes" % (len(pathes)-remaining, len(pathes), num_processes))
-
-        res = pool.map(PXDParser.parse_pxd_file, args)
+        args = [full_paths.pop() for k in range(n_work - remaining)]
+        logger.log(
+            25,
+            "parsing progress %s out of %s with %s processes"
+            % (len(paths) - remaining, len(paths), num_processes),
+        )
+        parse_pxd_file_warn = partial(
+            PXDParser.parse_pxd_file, warn_level=cython_warn_level
+        )
+        res = pool.map(parse_pxd_file_warn, args)
         for r in res:
             decls.extend(r)
 
-    return _resolve_decls(decls)  
+    return _resolve_decls(decls)
 
 
 def resolve_decls_from_string(pxd_in_a_string):
@@ -272,7 +311,7 @@ def resolve_decls_from_string(pxd_in_a_string):
 def _resolve_decls(decls):
     """
     input:
-        class_decls ist list of instances of PXDParser.BaseDecl.
+        class_decls is a list of instances of PXDParser.BaseDecl.
         (contains annotations
           - about instance names for template parameterized classes
           - about inheritance of methods from other classes in class_decls
@@ -281,7 +320,7 @@ def _resolve_decls(decls):
         list of instances of ResolvedClass
     """
 
-    L.info("start resolving decls")
+    logger.info("start resolving decls")
     assert all(isinstance(d, PXDParser.BaseDecl) for d in decls)
 
     def filter_out(tt):
@@ -301,22 +340,28 @@ def _resolve_decls(decls):
     for e in enum_decls:
         enum_mapping[e.name] = Types.CppType(e.name, enum_items=e.items)
 
-    # register class aliases
+    # register class aliases, e.g. instantiated template classes
     instance_mapping = _parse_all_wrap_instances_comments(class_decls)
+
     # remove local targ mappings:
     instance_mapping = dict((k, v0) for (k, (v0, v1)) in instance_mapping.items())
 
-    # resolve  typedefs in class aliase values
-    instance_mapping = dict((k, v.transformed(typedef_mapping)) for (k, v) in
-                            instance_mapping.items())
+    # resolve typedefs in class alias values
+    instance_mapping = dict(
+        (k, v.transformed(typedef_mapping)) for (k, v) in instance_mapping.items()
+    )
 
     # add enum mapping to class instances mapping
     intersecting_names = set(instance_mapping) & set(enum_mapping)
-    assert not intersecting_names, "enum names and class decls overlap: %s" % intersecting_names
+    assert not intersecting_names, (
+        "enum names and class decls overlap: %s" % intersecting_names
+    )
 
     instance_mapping.update(enum_mapping)
 
-    functions = [_resolve_function(f, instance_mapping, typedef_mapping) for f in function_decls]
+    functions = [
+        _resolve_function(f, instance_mapping, typedef_mapping) for f in function_decls
+    ]
 
     enums = [ResolvedEnum(e) for e in enum_decls]
     typedefs = [ResolvedTypeDef(t) for t in typedef_decls]
@@ -367,7 +412,7 @@ def _generate_inheritance_graph(class_decls, name_to_decl):
         for base_decl_str in cdcl.annotations.get("wrap-inherits", []):
             base = Types.CppType.from_string(base_decl_str)
             base_class = name_to_decl[base.base_type]
-            L.info("%s inherits %s" % (cdcl.name, base))
+            logger.info("%s inherits %s" % (cdcl.name, base))
             graph[cdcl].append((base_class, base.template_args))
     return graph
 
@@ -386,9 +431,9 @@ def _resolve_inheritance(cdcl, class_decls, inheritance_graph):
     that is: methods from super_classes and their super_classes.
     """
 
-    L.info("resolve_inheritance for %s" % cdcl.name)
+    logger.info("resolve_inheritance for %s" % cdcl.name)
 
-    # first we recurses to all super classes:
+    # first we recurse over all super classes:
     for super_cld, _ in inheritance_graph[cdcl]:
         _resolve_inheritance(super_cld, class_decls, inheritance_graph)
 
@@ -400,28 +445,31 @@ def _resolve_inheritance(cdcl, class_decls, inheritance_graph):
 
 def _add_inherited_methods(cdcl, super_cld, used_parameters):
 
-    L.info("add_inherited_methods for %s" % cdcl.name)
+    logger.info("add_inherited_methods for %s" % cdcl.name)
 
     super_targs = super_cld.template_parameters
-    # template paremeer None behaves like []
+    # template parameter None behaves like []
     used_parameters = used_parameters or []
     super_targs = super_targs or []
 
-    # check if parmetirization signature matches:
+    # check if parameterization signature matches:
     if len(used_parameters) != len(super_targs):
-        raise Exception("deriving %s from %s does not match" % (cdcl.name, super_cld.name))
+        raise Exception(
+            "deriving %s from %s does not match" % (cdcl.name, super_cld.name)
+        )
 
     # map template parameters in super class to the parameters used in current
     # class:
     mapping = dict(zip(super_targs, used_parameters))
-    # get copy of methods from super class ans transform template params:
+    # get copy of methods from super class and transform template params:
     transformed_methods = super_cld.get_transformed_methods(mapping)
-    transformed_methods = dict((k, v) for (k, v) in transformed_methods.items()
-                               if k != super_cld.name)  # remove constructors
+    transformed_methods = dict(
+        (k, v) for (k, v) in transformed_methods.items() if k != super_cld.name
+    )  # remove constructors
     for method in transformed_methods:
-        L.info("attach to %s: %s" % (cdcl.name, method))
+        logger.info("attach to %s: %s" % (cdcl.name, method))
     cdcl.attach_base_methods(transformed_methods)
-    # L.info("")
+    # logger.info("")
 
 
 def _build_typedef_mapping(decls):
@@ -439,16 +487,16 @@ def _check_typedefs(decls):
         raise Exception(msg)
 
 
-def _parse_all_wrap_instances_comments(class_decls):
-    """ parses annotations of all classes and registers aliases for
-        classes.
+def _parse_all_wrap_instances_comments(class_decls: List[PXDParser.CppClassDecl]):
+    """parses annotations of all classes and registers aliases for
+    classes.
 
-        cdef cppclass A[U]:
-            #wrap-instances:
-            #  AA := A[int]
+    cdef cppclass A[U]:
+        #wrap-instances:
+        #  AA := A[int]
 
-        generates an entry  'AA' : ( A[int], {'U': 'int'} ) in r
-        where "A[int]" is CppType("A", [CppType["int"])
+    generates an entry  'AA' : ( A[int], {'U': 'int'} ) in r
+    where "A[int]" is CppType("A", [CppType["int"])
 
     """
     r = dict()
@@ -457,7 +505,9 @@ def _parse_all_wrap_instances_comments(class_decls):
     return r
 
 
-def _parse_wrap_instances_comments(cdcl):
+def _parse_wrap_instances_comments(
+    cdcl: PXDParser.CppClassDecl,
+) -> Dict[AnyStr, Tuple[Types.CppType, Dict]]:
 
     inst_annotations = cdcl.annotations.get("wrap-instances")
     r = dict()
@@ -472,7 +522,7 @@ def _parse_wrap_instances_comments(cdcl):
     return r
 
 
-def parse_alias(cdcl, instance_decl_str):
+def parse_alias(cdcl: PXDParser.CppClassDecl, instance_decl_str: AnyStr):
     """
     instance_decl_str looks like "Tint := T[int]"
     """
@@ -488,7 +538,7 @@ def parse_alias(cdcl, instance_decl_str):
     return name, type_, t_param_mapping
 
 
-def parse_inst_decl(str_):
+def parse_inst_decl(str_: AnyStr) -> Tuple[AnyStr, Types.CppType]:
     """
     instance_decl_str looks like "Tint := T[int]"
     returns "Tint", CppType.from_string("T[int]")
@@ -497,16 +547,17 @@ def parse_inst_decl(str_):
         left, __, right = str_.partition(":=")
         name, decl_str = left.strip(), right.strip()
         return name, Types.CppType.from_string(decl_str)
-    except:
-        raise Exception("could not parse instance delcaration '%s'" % str_)
+    except ValueError:
+        raise Exception("could not parse instance declaration '%s'" % str_)
 
 
 def _resolve_class_decls(class_decls, typedef_mapping, instance_mapping):
-    """
-    """
+    """ """
     all_resolved_classes = []
     for class_decl in class_decls:
-        resolved_classes = _resolve_class_decl(class_decl, typedef_mapping, instance_mapping)
+        resolved_classes = _resolve_class_decl(
+            class_decl, typedef_mapping, instance_mapping
+        )
         all_resolved_classes.extend(resolved_classes)
     return all_resolved_classes
 
@@ -514,7 +565,7 @@ def _resolve_class_decls(class_decls, typedef_mapping, instance_mapping):
 def _resolve_class_decl(class_decl, typedef_mapping, i_mapping):
     # one decl can produce multiple classes !
 
-    L.info("resolve class decl %s" % class_decl.name)
+    logger.info("resolve class decl %s" % class_decl.name)
 
     r = _parse_wrap_instances_comments(class_decl)
     resolved_classes = []
@@ -532,19 +583,24 @@ def _resolve_class_decl(class_decl, typedef_mapping, i_mapping):
                 if ignore:
                     continue
                 if mdcl.name == class_decl.name:
-                    r_method = _resolve_constructor(cinst_name, mdcl, i_mapping, local_mapping)
+                    r_method = _resolve_constructor(
+                        cinst_name, mdcl, i_mapping, local_mapping
+                    )
                 else:
                     r_method = _resolve_method(mdcl, i_mapping, local_mapping)
                 r_methods.append(r_method)
-        r_class = ResolvedClass(cinst_name, r_methods, r_attributes,
-                                class_decl, i_mapping, local_mapping)
+        r_class = ResolvedClass(
+            cinst_name, r_methods, r_attributes, class_decl, i_mapping, local_mapping
+        )
         resolved_classes.append(r_class)
     return resolved_classes
 
 
 def _build_local_typemap(t_param_mapping, typedef_mapping):
-    # for resolving typedefed types in template instance args:
-    local_map = dict((n, t.transformed(typedef_mapping)) for (n, t) in t_param_mapping.items())
+    # for resolving typedef'ed types in template instance args:
+    local_map = dict(
+        (n, t.transformed(typedef_mapping)) for (n, t) in t_param_mapping.items()
+    )
 
     # for resolving 'free' typedefs in method args and result types:
     if set(local_map) & set(typedef_mapping):
@@ -556,46 +612,58 @@ def _build_local_typemap(t_param_mapping, typedef_mapping):
 
 
 def _resolve_constructor(cinst_name, method_decl, instance_mapping, local_type_map):
-    L.info("resolve method decl: '%s'" % method_decl)
-    result = _resolve_method_or_function(method_decl, instance_mapping,
-                                         local_type_map, ResolvedMethod)
+    logger.info("resolve method decl: '%s'" % method_decl)
+    result = _resolve_method_or_function(
+        method_decl, instance_mapping, local_type_map, ResolvedMethod
+    )
     result.name = cinst_name
-    # L.info("result             : '%s'" % result)
-    # L.info("")
+    # logger.info("result             : '%s'" % result)
+    # logger.info("")
     return result
 
 
 def _resolve_method(method_decl, instance_mapping, local_type_map):
-    L.info("resolve method decl: '%s'" % method_decl)
-    result = _resolve_method_or_function(method_decl, instance_mapping,
-                                         local_type_map, ResolvedMethod)
-    # L.info("result             : '%s'" % result)
-    # L.info("")
+    logger.info("resolve method decl: '%s'" % method_decl)
+    result = _resolve_method_or_function(
+        method_decl, instance_mapping, local_type_map, ResolvedMethod
+    )
+    # logger.info("result             : '%s'" % result)
+    # logger.info("")
     return result
 
 
 def _resolve_function(method_decl, instance_mapping, local_type_map):
-    L.info("resolve function decl: '%s'" % method_decl)
-    result = _resolve_method_or_function(method_decl, instance_mapping,
-                                         local_type_map, ResolvedFunction)
-    # L.info("result               : '%s'" % result)
-    # L.info("")
+    logger.info("resolve function decl: '%s'" % method_decl)
+    result = _resolve_method_or_function(
+        method_decl, instance_mapping, local_type_map, ResolvedFunction
+    )
+    # logger.info("result               : '%s'" % result)
+    # logger.info("")
     return result
 
 
-def _resolve_method_or_function(method_decl, instance_mapping, local_type_map,
-                                clz):
+def _resolve_method_or_function(method_decl, instance_mapping, local_type_map, clz):
     """
     resolves aliases in return and argument types
     """
-    result_type = _resolve_alias(method_decl.result_type, instance_mapping,
-                                 local_type_map)
+    result_type = _resolve_alias(
+        method_decl.result_type, instance_mapping, local_type_map
+    )
     args = []
     for arg_name, arg_type in method_decl.arguments:
         arg_type = _resolve_alias(arg_type, instance_mapping, local_type_map)
         args.append((arg_name, arg_type))
     name = method_decl.annotations.get("wrap-as", method_decl.name)
-    return clz(name, result_type, args, method_decl, instance_mapping, local_type_map)
+
+    return clz(
+        name,
+        method_decl.is_static,
+        result_type,
+        args,
+        method_decl,
+        instance_mapping,
+        local_type_map,
+    )
 
 
 def _resolve_attribute(adecl, instance_mapping, type_map):
