@@ -1,6 +1,5 @@
 # encoding: utf-8
-from __future__ import print_function
-from __future__ import absolute_import
+from __future__ import print_function, absolute_import, annotations
 
 __license__ = """
 
@@ -32,37 +31,57 @@ OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 """
 
+from typing import Dict, Union, List, Collection, Sequence, AnyStr, TYPE_CHECKING
+
+if TYPE_CHECKING:
+    from autowrap.DeclResolver import ResolvedMethod
+
+import Cython.Compiler.Nodes
 from Cython.Compiler.CmdLine import parse_command_line
-from Cython.Compiler.Main import create_default_resultobj, CompilationSource
-from Cython.Compiler import Pipeline
+from Cython.Compiler.Main import create_default_resultobj, CompilationSource, Context
+from Cython.Compiler import Pipeline, FusedNode
 from Cython.Compiler.Scanning import FileSourceDescriptor
 from Cython.Compiler.Nodes import *
 from Cython.Compiler.ExprNodes import *
+import Cython.Compiler.Errors
 
 from autowrap.Types import CppType
 
 import os
-import sys
 
-# import logging as L
+from autowrap import logger
+from autowrap.Code import Code as Code
 
 from collections import defaultdict
 from .tools import OrderKeepingDictionary
 
+AnnotDict = Dict[str, Union[bool, List[str]]]
+
 """
 Methods in this module use Cythons Parser to build an Cython syntax tree
-from the annotated .pxd files and creates a represenation of the
+from the annotated .pxd files and creates a representation of the
 included classes and methods.
 """
 
 
-def parse_class_annotations(node, lines):
-    """ parses wrap-instructions inside comments below class def """
+def _check_type_constness(ctype: Nodes.CBaseTypeNode) -> bool:
+    """Regardless of Cython version, checks whether the passed Cython type is const"""
+    try:
+        return isinstance(ctype, Nodes.CConstTypeNode)
+    except AttributeError:
+        return isinstance(ctype, Nodes.CConstOrVolatileTypeNode) and ctype.is_const
+
+
+def parse_class_annotations(node, lines: Collection[str]) -> AnnotDict:
+    """parses wrap-instructions inside comments below class def.
+    Either returns a list of strings for annotations/directives with a ':' and following strings
+    or True/False for boolean directives without ':'
+    """
     start_at_line = node.pos[1]
     return _parse_multiline_annotations(lines[start_at_line:])
 
 
-def _parse_multiline_annotations(lines):
+def _parse_multiline_annotations(lines: Collection[str]) -> AnnotDict:
     """does the hard work for parse_class_annotations, and is
     better testable than this.
     """
@@ -81,18 +100,41 @@ def _parse_multiline_annotations(lines):
                 key = line.rstrip(":")
                 line = next(it).strip()
                 while line.startswith("#  "):
-                    value = line[1:].strip()
-                    result[key].append(value)
-                    line = next(it).strip()
+                    if key == "wrap-doc":
+                        value = line[3:].rstrip()  # rstrip to keep indentation in docs
+                    else:
+                        value = line[1:].strip()
+
+                    if (
+                        key == "wrap-doc" or value
+                    ):  # don't add empty non wrap-doc values
+                        result[key].append(value)
+
+                    try:
+                        line = next(it).lstrip()  # lstrip to keep empty lines in docs
+                    except StopIteration:
+                        break
             else:
                 key = line
                 result[key] = True
         else:
             break
+
+    # make sure wrap-doc is always a Code object
+    if "wrap-doc" in result.keys():
+        doc = result.get("wrap-doc", [])
+        if isinstance(doc, basestring):
+            doc = [doc]
+
+        c = Code()
+        c.addRawList(doc)
+        result["wrap-doc"] = c
     return result
 
 
-def parse_line_annotations(node, lines):
+def parse_line_annotations(
+    node: Cython.Compiler.Nodes.Node, lines: Sequence[str]
+) -> AnnotDict:
     """
     parses comments at end of line, in most cases the lines of
     method declarations.
@@ -113,29 +155,49 @@ def parse_line_annotations(node, lines):
         break
 
     for line in lines[start:end]:
-        __, __, comment = line.partition("#")
-        if comment:
-            key = None
-            for f_ in comment.split(" "):
-                f = f_.strip()
-                if not f:
-                    continue
-                if ":" in f:
-                    key, value = f.split(":", 1)
-                    assert value.strip(), "empty value for key '%s'" % key
-                    result[key] = value
-                elif f.find("wrap-") != -1:
-                    key, value = f, True
-                    result[key] = value
-                elif key is not None:
-                    # they belong to the previous key
-                    value = " " + f_
-                    result[key] += value
+        try:
+            __, __, comment = line.partition("#")
+            if comment:
+                key = None
+                for f_ in comment.split(" "):
+                    f = f_.strip()
+                    if not f:
+                        continue
+                    if ":" in f:
+                        key, value = f.split(":", 1)
+                        assert value.strip(), (
+                            "empty value (or excess space?) for key '%s' in line '%s'"
+                            % (key, line.rstrip())
+                        )
+                        result[key] = value
+                    elif f.find("wrap-") != -1:
+                        key, value = f, True
+                        result[key] = value
+                    elif key is not None:
+                        # they belong to the previous key
+                        value = " " + f_
+                        result[key] += value
+        except Exception as e:
+            raise ValueError("Cannot parse '{}'".format(line)) from e
+    # check for multi line annotations after method declaration
+    additional_annotations = _parse_multiline_annotations(lines[end:])
+    # add multi line doc string to result (overwrites single line wrap-doc, if exists)
+    if "wrap-doc" in additional_annotations.keys():
+        result["wrap-doc"] = additional_annotations["wrap-doc"]
+    else:
+        # make sure wrap-doc is always a Code object
+        if "wrap-doc" in result.keys():
+            doc = result.get("wrap-doc", [])
+            if isinstance(doc, basestring):
+                doc = [doc]
+
+            c = Code()
+            c.addRawList(doc)
+            result["wrap-doc"] = c
     return result
 
 
-def _extract_template_args(node):
-
+def _extract_template_args(node: Cython.Compiler.Nodes.Node) -> CppType:
     if isinstance(node, NameNode):
         return CppType(node.name, None)
 
@@ -147,14 +209,19 @@ def _extract_template_args(node):
     elif isinstance(node.index, NameNode):
         args = [CppType(node.index.name)]
     else:
-        raise Exception("can not handle node %s in template arg decl" % node.index)
+        raise Exception(
+            "Can not handle node %s in template argument declaration" % node.index
+        )
     return CppType(name, args)
 
 
-def _extract_type(base_type, decl):
-    """ extracts type information from node in parse_pxd_file tree """
+def _extract_type(
+    base_type: Cython.Compiler.Nodes.CBaseTypeNode,
+    decl: Cython.Compiler.Nodes.CDeclaratorNode,
+) -> CppType:
+    """extracts type information from node in parse_pxd_file tree"""
 
-    type_is_const = isinstance(base_type, Nodes.CConstTypeNode)
+    type_is_const = _check_type_constness(base_type)
 
     # Complex const values, e.g. "const Int *" need to be reduced to base type
     # to get the correct name. Note: we will have to deal with const-ness first
@@ -182,7 +249,7 @@ def _extract_type(base_type, decl):
                 # Handle const template arguments which do not have a name
                 # themselves, only their base types have name attribute (see
                 # for example: shared_ptr<const Int>)
-                arg_is_const = isinstance(arg_node.base_type, Nodes.CConstTypeNode)
+                arg_is_const = _check_type_constness(arg_node.base_type)
                 if arg_is_const:
                     name = arg_node.base_type.base_type.name
                 else:
@@ -212,7 +279,12 @@ def _extract_type(base_type, decl):
                 tt = _extract_template_args(arg_node)
                 template_parameters.append(tt)
             else:
-                raise Exception("can not handle template arg_node %r" % arg_node)
+                raise Exception(
+                    "Can not handle template argument node (arg_node) %r"
+                    % arg_node.pos[0].file_path
+                    + " line: %r" % arg_node.pos[1]
+                    + " col: %r" % arg_node.pos[2]
+                )
 
         base_type = base_type.base_type_node
 
@@ -233,14 +305,16 @@ def _extract_type(base_type, decl):
 
 
 class BaseDecl(object):
-    def __init__(self, name, annotations, pxd_path):
-        self.name = name
-        self.annotations = annotations
-        self.pxd_path = pxd_path
+    def __init__(
+        self, name: str, annotations: Dict[str, Union[bool, List[str]]], pxd_path: str
+    ):
+        self.name: str = name
+        self.annotations: AnnotDict = annotations
+        self.pxd_path: str = pxd_path
 
 
 class CTypeDefDecl(BaseDecl):
-    def __init__(self, new_name, type_, annotations, pxd_path):
+    def __init__(self, new_name: str, type_, annotations: AnnotDict, pxd_path):
         super(CTypeDefDecl, self).__init__(new_name, annotations, pxd_path)
         self.type_ = type_
 
@@ -257,15 +331,20 @@ class CTypeDefDecl(BaseDecl):
 
 
 class EnumDecl(BaseDecl):
-    def __init__(self, name, items, annotations, pxd_path):
+    def __init__(self, name, scoped, items, annotations, pxd_path):
         super(EnumDecl, self).__init__(name, annotations, pxd_path)
+        self.scoped = scoped
         self.items = items
         self.template_parameters = None
 
     @classmethod
-    def parseTree(cls, node, lines, pxd_path):
+    def parseTree(cls, node: CEnumDefNode, lines, pxd_path):
         name = node.name
         items = []
+        try:
+            scoped = node.scoped
+        except AttributeError:
+            scoped = False
         annotations = parse_class_annotations(node, lines)
         current_value = 0
         for item in node.items:
@@ -274,10 +353,11 @@ class EnumDecl(BaseDecl):
             items.append((item.name, current_value))
             current_value += 1
 
-        return cls(name, items, annotations, pxd_path)
+        return cls(name, scoped, items, annotations, pxd_path)
 
     def __str__(self):
         res = "EnumDecl %s : " % self.name
+        res += "scoped" if self.scoped else ""
         res += ", ".join("%s: %d" % (i, v) for (i, v) in self.items)
         return res
 
@@ -286,6 +366,10 @@ class EnumDecl(BaseDecl):
 
 
 class CppClassDecl(BaseDecl):
+    methods: Dict[AnyStr, List[CppMethodOrFunctionDecl]]
+    attributes: List[CppAttributeDecl]
+    template_parameters: List[AnyStr]
+
     def __init__(
         self, name, template_parameters, methods, attributes, annotations, pxd_path
     ):
@@ -295,7 +379,9 @@ class CppClassDecl(BaseDecl):
         self.template_parameters = template_parameters
 
     @classmethod
-    def parseTree(cls, node, lines, pxd_path):
+    def parseTree(
+        cls, node: Cython.Compiler.Nodes.CppClassNode, lines: Collection[str], pxd_path
+    ):
         name = node.name
         template_parameters = node.templates
         if (
@@ -312,12 +398,37 @@ class CppClassDecl(BaseDecl):
         methods = OrderKeepingDictionary()
         attributes = []
         for att in node.attributes:
-            decl = MethodOrAttributeDecl.parseTree(att, lines, pxd_path)
+            decl = None
+            if isinstance(att, CVarDefNode):  # attribute or member function
+                decl = MethodOrAttributeDecl.parseTree(att, lines, pxd_path)
+            elif isinstance(att, CEnumDefNode):
+                logger.warning(
+                    "Nested enums currently not supported by autowrap. Skipping its wrap."
+                    " Please add them under a new cdef extern section with class namespace. E.g.: \n"
+                    "cdef extern from 'foo.hpp' namespace 'Foo': \n"
+                    "    cpdef enum class MyEnum 'Foo::MyEnum': \n"
+                    "      # wrap-attach: Foo \n"
+                    "      A,B,C"
+                )
+                # TODO we might be able to support it with the following
+                # decl = EnumDecl.parseTree(att, lines, pxd_path)
+            elif isinstance(att, CClassDefNode):
+                logger.warning(
+                    "Nested classes are currently not supported by autowrap. Skipping its wrap."
+                    " Try to add them under a new cdef extern section with class namespace. E.g.: \n"
+                    "cdef extern from 'foo.hpp' namespace 'Foo': \n"
+                    "    cpdef cppclass MyClass 'Foo::MyClass': \n"
+                    "      ..."
+                )
             if decl is not None:
                 if isinstance(decl, CppMethodOrFunctionDecl):
                     methods.setdefault(decl.name, []).append(decl)
                 elif isinstance(decl, CppAttributeDecl):
                     attributes.append(decl)
+                elif isinstance(decl, EnumDecl):
+                    # Should not happen since we currently forbid it in the logic above
+                    # attributes.append(decl)
+                    pass
 
         return cls(
             name, template_parameters, methods, attributes, class_annotations, pxd_path
@@ -354,22 +465,28 @@ class CppClassDecl(BaseDecl):
 
 
 class CppAttributeDecl(BaseDecl):
-    def __init__(self, name, type_, annotations, pxd_path):
+    def __init__(self, name, type_: CppType, annotations, pxd_path):
         super(CppAttributeDecl, self).__init__(name, annotations, pxd_path)
         self.type_ = type_
 
 
 class CppMethodOrFunctionDecl(BaseDecl):
-    def __init__(self, result_type, name, arguments, annotations, pxd_path):
+    def __init__(self, result_type, name, arguments, is_static, annotations, pxd_path):
         super(CppMethodOrFunctionDecl, self).__init__(name, annotations, pxd_path)
         self.result_type = result_type
         self.arguments = arguments
+        self.is_static = is_static
 
     def transformed(self, typemap):
         result_type = self.result_type.transformed(typemap)
         args = [(n, t.transformed(typemap)) for n, t in self.arguments]
         return CppMethodOrFunctionDecl(
-            result_type, self.name, args, self.annotations, self.pxd_path
+            result_type,
+            self.name,
+            args,
+            self.is_static,
+            self.annotations,
+            self.pxd_path,
         )
 
     def matches(self, other):
@@ -391,10 +508,15 @@ class CppMethodOrFunctionDecl(BaseDecl):
 
 class MethodOrAttributeDecl(object):
     @classmethod
-    def parseTree(cls, node, lines, pxd_path):
+    def parseTree(cls, node: CVarDefNode, lines, pxd_path):
+        if isinstance(node, CEnumDefNode):
+            return EnumDecl.parseTree(node, lines, pxd_path)
+
         annotations = parse_line_annotations(node, lines)
         if isinstance(node, CppClassNode):
-            return None  # nested classes only can be delcared in pxd
+            return None  # nested classes only can be declared in pxd
+
+        is_static = False
 
         (decl,) = node.declarators
         result_type = _extract_type(node.base_type, decl)
@@ -408,9 +530,17 @@ class MethodOrAttributeDecl(object):
         ):
             # Handle raw pointer declarations (call with base name)
             return CppAttributeDecl(decl.base.name, result_type, annotations, pxd_path)
+
+        # if the variable name declaration is part of a function declaration
         if isinstance(decl.base, CFuncDeclaratorNode):
+            # go up the base type
             decl = decl.base
 
+        if node.decorators is not None:
+            for dec in node.decorators:
+                if dec.decorator.name == "staticmethod":
+                    is_static = True
+
         name = decl.base.name
         args = []
         for arg in decl.args:
@@ -420,9 +550,10 @@ class MethodOrAttributeDecl(object):
             elif isinstance(argdecl, CPtrDeclaratorNode):
                 base = argdecl.base
                 if isinstance(base, CPtrDeclaratorNode):
-                    raise Exception("multi ptr not supported")
+                    raise Exception("multi ptr not supported for method " + name)
                 argname = base.name
             else:
+                # TODO document in which case the following if-case occurs and why two times .base
                 if not hasattr(argdecl, "name"):
                     argname = argdecl.base.base.name
                 else:
@@ -430,13 +561,9 @@ class MethodOrAttributeDecl(object):
             tt = _extract_type(arg.base_type, argdecl)
             args.append((argname, tt))
 
-        return CppMethodOrFunctionDecl(result_type, name, args, annotations, pxd_path)
-
-    def __str__(self):
-        rv = str(self.result_type)
-        rv += " " + self.name
-        argl = [str(type_) + " " + str(name) for name, type_ in self.arguments]
-        return rv + "(" + ", ".join(argl) + ")"
+        return CppMethodOrFunctionDecl(
+            result_type, name, args, is_static, annotations, pxd_path
+        )
 
 
 def parse_str(what):
@@ -451,7 +578,8 @@ def parse_str(what):
         return result
 
 
-def parse_pxd_file(path):
+def parse_pxd_file(path, warn_level=1):
+    Cython.Compiler.Errors.LEVEL = warn_level
 
     options, sources = parse_command_line(["--cplus", path])
 
@@ -470,7 +598,10 @@ def parse_pxd_file(path):
     source = CompilationSource(source_desc, name, os.getcwd())
     result = create_default_resultobj(source, options)
 
-    context = options.create_context()
+    try:  # Cython 0.X
+        context = options.create_context()
+    except:  # Cython 3.X
+        context = Context.from_options(options)
     pipeline = Pipeline.create_pyx_pipeline(context, options, result)
     context.setup_errors(options, result)
     root = pipeline[0](source)  # only parser
@@ -495,7 +626,7 @@ def parse_pxd_file(path):
             body = tree.body.body
             yield body
         else:
-            raise Exception("parse_pxd_file failed: no valied .pxd file !")
+            raise Exception("parse_pxd_file failed: no valid .pxd file !")
 
     lines = open(path).readlines()
 
@@ -514,7 +645,8 @@ def parse_pxd_file(path):
     for body in iter_bodies(root):
         handler = handlers.get(type(body))
         if handler is not None:
-            # L.info("parsed %s, handler=%s" % (body.__class__, handler.im_self))
+            # with open('log.txt', 'a') as f:
+            #    f.write("parsed %s, handler=%s \n" % (body.__class__, handler.__self__))
             result.append(handler(body, lines, path))
         else:
             for node in getattr(body, "stats", []):
@@ -528,4 +660,7 @@ if __name__ == "__main__":
 
     import sys
 
-    print(parse_pxd_file(sys.argv[1]))
+    if len(sys.argv) == 3:
+        print(parse_pxd_file(sys.argv[1], sys.argv[2]))
+    else:
+        print(parse_pxd_file(sys.argv[1]))
