@@ -9,6 +9,35 @@ template<class A> void _iadd(A * a1, const A * a2)
     (*a1) += (*a2);
 }
 
+template<class A> void _isub(A * a1, const A * a2)
+{
+    (*a1) -= (*a2);
+}
+
+template<class A> void _imul(A * a1, const A * a2)
+{
+    (*a1) *= (*a2);
+}
+
+template<class A> void _itruediv(A * a1, const A * a2)
+{
+    (*a1) /= (*a2);
+}
+
+template<class A> void _imod(A * a1, const A * a2)
+{
+    (*a1) %= (*a2);
+}
+
+template<class A> void _ilshift(A * a1, const A * a2)
+{
+    (*a1) <<= (*a2);
+}
+
+template<class A> void _irshift(A * a1, const A * a2)
+{
+    (*a1) >>= (*a2);
+}
 namespace autowrap {
 
     template <class X>
@@ -20,7 +49,7 @@ namespace autowrap {
 
         public:
 
-            AutowrapRefHolder(X &ref): _ref(ref) 
+            AutowrapRefHolder(X &ref): _ref(ref)
             {
             }
 
@@ -46,7 +75,7 @@ namespace autowrap {
 
             AutowrapPtrHolder() {}
 
-            AutowrapPtrHolder(X *ref): _ptr(ref) 
+            AutowrapPtrHolder(X *ref): _ptr(ref)
             {
             }
 
@@ -72,7 +101,7 @@ namespace autowrap {
 
             AutowrapConstPtrHolder() {}
 
-            AutowrapConstPtrHolder(const X *ref): _ptr(ref) 
+            AutowrapConstPtrHolder(const X *ref): _ptr(ref)
             {
             }
 
